package megacycles;

public class AffineTransform {

  public static AffineTransform getIdentityTransform() {
    return new AffineTransform(
      new double[][] {
        { 1, 0, 0, 0 },
        { 0, 1, 0, 0 },
        { 0, 0, 1, 0 },
        { 0, 0, 0, 1 },
      });
  }

  public static AffineTransform getTranslationTransform(
      double tx, double ty, double tz) {
    return new AffineTransform(
      new double[][] {
        { 1, 0, 0, tx },
        { 0, 1, 0, ty },
        { 0, 0, 1, tz },
        { 0, 0, 0, 1 },
      });
  }

  public static AffineTransform getScaleTransform(
      double sx, double sy, double sz) {
    return new AffineTransform(
      new double[][] {
        { sx, 0,  0, 0 },
        { 0, sy,  0, 0 },
        { 0,  0, sz, 0 },
        { 0,  0,  0, 1 },
      });
  }

  public static AffineTransform getRotateAroundXTransform(double angle) {
    double cos = Math.cos(angle);
    double sin = Math.sin(angle);
    return new AffineTransform(
      new double[][] {
        { 1,   0,    0, 0 },
        { 0, cos, -sin, 0 },
        { 0, sin,  cos, 0 },
        { 0,   0,    0, 1 },
      });
  }

  public static AffineTransform getRotateAroundYTransform(double angle) {
    double cos = Math.cos(angle);
    double sin = Math.sin(angle);
    return new AffineTransform(
      new double[][] {
        {  cos, 0, sin, 0 },
        {    0, 1,   0, 0 },
        { -sin, 0, cos, 0 },
        {    0, 0,   0, 1 },
      });
  }

  public static AffineTransform getRotateAroundZTransform(double angle) {
    double cos = Math.cos(angle);
    double sin = Math.sin(angle);
    return new AffineTransform(
      new double[][] {
        { cos, -sin, 0, 0 },
        { sin,  cos, 0, 0 },
        {   0,    0, 1, 0 },
        {   0,    0, 0, 1 },
      });
  }

  public double[][] matrix = new double[4][4];
  public double[][] original = new double[4][4];

  private static double[][] translateMatrix =  {
    { 1, 0, 0, 0 },
    { 0, 1, 0, 0 },
    { 0, 0, 1, 0 },
    { 0, 0, 0, 1 },
  };

  public AffineTransform() {    
  }

  public void translate(double x, double y, double z, boolean inverse) {
    for(int i = 0; i < 4; i++) {
      for(int j = 0; j < 4; j++) {
        original[i][j] = matrix[i][j];
      }
    }
    if (inverse) {
      translateMatrix[0][3] = x;
      translateMatrix[1][3] = y;
      translateMatrix[2][3] = z;
      for(int row = 0; row < 4; row++) {
        for(int col = 0; col < 4; col++) {
          double sum = 0;
          for(int i = 0; i < 4; i++) {
            sum += original[row][i] * translateMatrix[i][col];
          }
          matrix[row][col] = sum;
        }
      }
    } else {
      matrix[0][3] += x;
      matrix[1][3] += y;
      matrix[2][3] += z;
    }
  }

  public void undoTranslate() {
    double[][] temp = matrix;
    matrix = original;
    original = temp;
  }

  public void setTranslation(double x, double y, double z) {
    matrix[0][3] = x;
    matrix[1][3] = y;
    matrix[2][3] = z;
  }

  public AffineTransform(double[][] matrix) {
    for(int row = 0; row < 4; row++) {
      for(int col = 0; col < 4; col++) {
        this.matrix[row][col] = matrix[row][col];
      }
    }
  }

  public AffineTransform multiply(AffineTransform affineTransform) {
    AffineTransform product = new AffineTransform();
    double[][] matrix2 = affineTransform.matrix;
    double[][] result = product.matrix;
    for(int row = 0; row < 4; row++) {
      for(int col = 0; col < 4; col++) {
        double sum = 0;
        for(int i = 0; i < 4; i++) {
          sum += matrix[row][i] * matrix2[i][col];
        }
        result[row][col] = sum;
      }
    }
    return product;
  }

  public Vector multiply(Vector vector) {
    Vector product = new Vector();
    double[] vec = vector.vector;
    double[] result = product.vector;
    for(int row = 0; row < 4; row++) {
      double sum = 0;
      for(int col = 0; col < 4; col++) {
        sum += matrix[row][col] * vec[col];
      }
      result[row] = sum;
    }
    return product;
  }


  private Vector _product = new Vector();
  public void multiplyVector(Vector vector) {
    double[] vec = vector.vector;
    double[] result = _product.vector;
    for(int row = 0; row < 4; row++) {
      double sum = 0;
      for(int col = 0; col < 4; col++) {
        sum += matrix[row][col] * vec[col];
      }
      result[row] = sum;
    }
    _product.copyTo(vector);
  }
}
