package megacycles;

public class HilbertCurve {

  public static final byte EMPTY = 0;
  public static final byte LEFT = 1;
  public static final byte UP = 2;
  public static final byte RIGHT = 3;
  public static final byte DOWN = 4;

  private HilbertCurve() {
  }

  public static byte[][] generate(int order) {
    int side = 3 * (1 << order) - 2;
    byte[][] curve = new byte[side][side];

    StringBuilder sb = new StringBuilder("L");
    for(int i = 0; i < order; i++) {
      StringBuilder sb2 = new StringBuilder();
      for(int j = 0; j < sb.length(); j++) {
        char ch = sb.charAt(j);
        switch(ch) {
          case 'L':
            sb2.append("+RF-LFL-FR+");
            break;
          case 'R':
            sb2.append("-LF+RFR+FL-");
            break;
          default:
            sb2.append(ch);
            break;
        }
      }
      sb = sb2;
    }

    byte direction = RIGHT;
    byte lastDirection = direction;
    int x = 0;
    int y = side - 1;
    for(int i = 0; i < sb.length(); i++) {
      switch(sb.charAt(i)) {
        case '+':
          switch(direction) {
            case LEFT:
              direction = DOWN;
              break;
            case UP:
              direction = LEFT;
              break;
            case RIGHT:
              direction = UP;
              break;
            case DOWN:
              direction = RIGHT;
              break;
          }
          break;
        case '-':
          switch(direction) {
            case LEFT:
              direction = UP;
              break;
            case UP:
              direction = RIGHT;
              break;
            case RIGHT:
              direction = DOWN;
              break;
            case DOWN:
              direction = LEFT;
              break;
          }
          break;
        case 'F':
          curve[y][x] = (byte)((lastDirection << 3) | direction);
          lastDirection = direction;
          for(int j = 0; j < 3; j++) {
            switch(direction) {
              case LEFT:
                x--;
                break;
              case UP:
                y--;
                break;
              case RIGHT:
                x++;
                break;
              case DOWN:
                y++;
                break;
            }
            curve[y][x] = direction;
          }
          break;
      }
    }
    curve[y][x] = (byte)((lastDirection << 3) | direction);
    lastDirection = direction;

    /*for(int i = 0; i < side; i++) {
      for(int j = 0; j < side; j++) {
        switch(curve[i][j] & 7) {
          case LEFT:
            System.out.print("L");
            break;
          case UP:
            System.out.print("U");
            break;
          case RIGHT:
            System.out.print("R");
            break;
          case DOWN:
            System.out.print("D");
            break;
          default:
            System.out.print(".");
            break;
        }
      }
      System.out.println();
    }*/

    return curve;
  }
}
