package megacycles;

import java.awt.image.*;
import javax.imageio.*;
import java.io.*;
import java.util.Random;

public class Main {

  // HD = 1280 x 720 !!!

  public static final int WIDTH = 1280;
  public static final int HEIGHT = 720;
  //public static final int WIDTH = 640;
  //public static final int HEIGHT = 360;
  public static final int VIEWER = 640;
  public static final int SAMPLES = 5;
  //public static final int SAMPLES = 1;
  public static final double SQUARE_SIZE = 400;
  public static final double WHEEL_RADIUS = SQUARE_SIZE / (2.0 * Math.PI);
  public static final double GROUND_Y = 20 * SQUARE_SIZE;
  public static double LIGHT_X = 0;
  public static double LIGHT_Y = -10000;
  public static double LIGHT_Z = 10000;
  public static final double RIDER_Y = -WHEEL_RADIUS * 9;
  public static final double SKY_Y = -25 * SQUARE_SIZE;

  private void getPixel(double x, double y, double[][][] map, 
      double[] pixel, double[][] temp) {
    int x1 = (int)Math.floor(x);
    int y1 = (int)Math.floor(y);
    double percentX = x - x1;
    double percentY = y - y1;
    getPixel(x1, y1, map, temp[0]);
    getPixel(x1 + 1, y1, map, temp[1]);
    getPixel(x1, y1 + 1, map, temp[2]);
    getPixel(x1 + 1, y1 + 1, map, temp[3]);
    for(int i = 0; i < 3; i++) {
      double left = (temp[2][i] - temp[0][i]) * percentY + temp[0][i];
      double right = (temp[3][i] - temp[1][i]) * percentY + temp[1][i];
      pixel[i] = (right - left) * percentX + left;
    }
  }

  private void getPixel(int x, int y, double[][][] map, double[] pixel) {
    if (x < 0) {
      x = -x;
    }
    if (y < 0) {
      y = -y;
    }
    x &= 1023;
    y &= 1023;
    if (x > 511) {
      x = 1023 - x;
    }
    if (y > 511) {
      y = 1023 - y;
    }
    pixel[0] = map[y][x][0];
    pixel[1] = map[y][x][1];
    pixel[2] = map[y][x][2];
  }

  private void generateLandscape(double[][][] pixels, boolean sky) {

    Random random = new Random(1);
    int[][] intensities = new int[512][512];
    int[] total = new int[2];

    total[0] = 0;
    total[1] = 1;
    setIntensity(intensities, 0, 0, total, random.nextInt(255));
    setIntensity(intensities, 511, 0, total, random.nextInt(255));
    setIntensity(intensities, 0, 511, total, random.nextInt(255));
    setIntensity(intensities, 511, 511, total, random.nextInt(255));
    generateLandscape(intensities, 0, 0, 511, 511, random, total);

    int[][] palette = new int[256][3];
    if (sky) {
      // blue sky: 6C 9B D8
      // cloud: EF EF EF
      for(int i = 0; i < 256; i++) {
        double percent = i / 255.0;
        percent = Math.max(0, Math.min(1, (0.5 + Math.atan(10.0 * (percent - .9)) / Math.PI)));
        palette[i][0] = (int)(0x6C + (0xEF - 0x6C) * percent);
        palette[i][1] = (int)(0x9B + (0xEF - 0x9B) * percent);
        palette[i][2] = (int)(0xD8 + (0xEF - 0xD8) * percent);
      }
    } else {
      // earth: A8 99 4F
      //water: 51 8E BB
      for(int i = 0; i < 256; i++) {
        double percent = i / 255.0;
        percent = Math.max(0, Math.min(1, (0.5 + Math.atan(50.0 * (percent - .5)) / Math.PI)));
        palette[i][0] = (int)(0xA8 + (0x51 - 0xA8) * percent);
        palette[i][1] = (int)(0x99 + (0x8E - 0x99) * percent);
        palette[i][2] = (int)(0x4F + (0xBB - 0x4F) * percent);
      }
    }

    for(int i = 0; i < 512; i++) {
      for(int j = 0; j < 512; j++) {
        int[] color = palette[intensities[i][j]];
        pixels[i][j][0] = color[0];
        pixels[i][j][1] = color[1];
        pixels[i][j][2] = color[2];
      }
    }
  }

  private void generateLandscape(
      int[][] intensities, int x1, int y1, int x2, int y2, Random random,
      int[] total) {

    if (x2 - x1 < 2 && y2 - y1 < 2) {
      return;
    }

    int midX = (x1 + x2) / 2;
    int midY = (y1 + y2) / 2;
    double percent = 10.0 * ((double)(x2 - x1)) / intensities.length;

    total[0] = total[1] = 0;
    addIntensity(intensities, x1, y1, total);
    addIntensity(intensities, x2, y1, total);
    addIntensity(intensities, x1, y2, total);
    addIntensity(intensities, x2, y2, total);
    setIntensity(intensities, midX, midY, total, 
        (int)(255 * percent * (random.nextDouble() - 0.5)));

    total[0] = total[1] = 0;
    addIntensity(intensities, x1, y1, total);
    addIntensity(intensities, x2, y1, total);
    addIntensity(intensities, midX, midY, total);
    setIntensity(intensities, midX, y1, total,
        (int)(255 * percent * (random.nextDouble() - 0.5)));

    total[0] = total[1] = 0;
    addIntensity(intensities, x1, y1, total);
    addIntensity(intensities, x1, y2, total);
    addIntensity(intensities, midX, midY, total);
    setIntensity(intensities, x1, midY, total,
        (int)(255 * percent * (random.nextDouble() - 0.5)));

    total[0] = total[1] = 0;
    addIntensity(intensities, x2, y1, total);
    addIntensity(intensities, x2, y2, total);
    addIntensity(intensities, midX, midY, total);
    setIntensity(intensities, x2, midY, total,
        (int)(255 * percent * (random.nextDouble() - 0.5)));

    total[0] = total[1] = 0;
    addIntensity(intensities, x1, y2, total);
    addIntensity(intensities, x2, y2, total);
    addIntensity(intensities, midX, midY, total);
    setIntensity(intensities, midX, y2, total,
        (int)(255 * percent * (random.nextDouble() - 0.5)));

    generateLandscape(intensities, x1, y1, midX, midY, random, total);
    generateLandscape(intensities, midX, y1, x2, midY, random, total);
    generateLandscape(intensities, x1, midY, midX, y2, random, total);
    generateLandscape(intensities, midX, midY, x2, y2, random, total);
  }

  private void setIntensity(int[][] intensities, int x, int y, int[] total, int random) {
    if (total[1] > 0) {
      int value = random + (total[0] / total[1]);
      if (value < 0) {
        value = 0;
      } else if (value > 255) {
        value = 255;
      }
      if (intensities[y][x] == 0) {
        intensities[y][x] = value;
      }
    }
  }

  private void addIntensity(int[][] intensities, int x, int y, int[] total) {
    if (x >= 0 && y >= 0 && x < intensities.length && y < intensities.length) {
      total[0] += intensities[y][x];
      total[1]++;
    }
  }

  public void render(int scene, int frame, String path) throws Throwable {

    boolean renderMen = true;

    double[][][] skyMap = new double[512][512][3];
    generateLandscape(skyMap, true);
    double[][][] earthMap = new double[512][512][3];
    generateLandscape(earthMap, false);
    double[] mapPixel = new double[3];
    double[][] tempMapPixels = new double[4][3];

    byte[][] hilbertCurve = HilbertCurve.generate(8);
    boolean[][] visited = new boolean[hilbertCurve.length][hilbertCurve.length];
    double[][][] pixels = new double[HEIGHT][WIDTH][3];

    Vector upperLeft = new Vector(-320, -240, 0);
    Vector upperRight = new Vector(320, -240, 0);
    Vector lowerLeft = new Vector(-320, 240, 0);
    Vector lowerRight = new Vector(320, 240, 0);
    Vector viewer = new Vector(0, 0, -VIEWER);
    Vector[] camera = { upperLeft, upperRight, lowerLeft, lowerRight, viewer };

    AffineTransform transform = null;

    if (scene == 1) {
      // SCENE 1
      LIGHT_X = 2500;
      LIGHT_Y = -500;
      LIGHT_Z = 9500;
      AffineTransform translation
          = AffineTransform.getTranslationTransform(1200, -1000, 8700);
      AffineTransform rotationY
          = AffineTransform.getRotateAroundYTransform(-Math.PI * 0.5 - 0.1);
      AffineTransform rotationX
          = AffineTransform.getRotateAroundXTransform(-0.7);
      transform = rotationY.multiply(rotationX);
      transform = translation.multiply(transform);
    } else if (scene == 2) {
      // SCENE 2
      LIGHT_X = 7000;
      LIGHT_Y = -8000;
      LIGHT_Z = 12000;
      AffineTransform translation
          = AffineTransform.getTranslationTransform(1900, -300, 10000);
      AffineTransform rotationY
          = AffineTransform.getRotateAroundYTransform(-Math.PI * 0.5 - 0.25);
      AffineTransform rotationX
          = AffineTransform.getRotateAroundXTransform(0.05);
      transform = rotationY.multiply(rotationX);
      transform = translation.multiply(transform);
    } else if (scene == 3) {
      // SCENE 3
      LIGHT_X = 2000;
      LIGHT_Y = -5000;
      LIGHT_Z = 9000;
      AffineTransform translation
          = AffineTransform.getTranslationTransform(1550, -3500, 9200);
      AffineTransform rotationY
          = AffineTransform.getRotateAroundYTransform(-Math.PI * 0.5);
      AffineTransform rotationX
          = AffineTransform.getRotateAroundXTransform(-Math.PI * 0.5 + 0.25);
      transform = rotationY.multiply(rotationX);
      transform = translation.multiply(transform);
    } else if (scene == 4) {
      // SCENE 4
      LIGHT_X = 4000;
      LIGHT_Y = -10000;
      LIGHT_Z = 3000;
      AffineTransform translation
          = AffineTransform.getTranslationTransform(2500, -5000, 5500);
      AffineTransform rotationY
          = AffineTransform.getRotateAroundYTransform(-Math.PI);
      AffineTransform rotationX
          = AffineTransform.getRotateAroundXTransform(-Math.PI * 0.5 + 0.6);
      transform = rotationY.multiply(rotationX);
      transform = translation.multiply(transform);
    } else if (scene == 5) {
      // SCENE 5
      LIGHT_X = 12500;
      LIGHT_Y = -12000;
      LIGHT_Z = 10000;
      AffineTransform translation
          = AffineTransform.getTranslationTransform(10500, -1000, 13600);
      AffineTransform rotationY
          = AffineTransform.getRotateAroundYTransform(0);
      AffineTransform rotationX
          = AffineTransform.getRotateAroundXTransform(-Math.PI * 0.5 + 1.05);
      transform = rotationY.multiply(rotationX);
      transform = translation.multiply(transform);
    } else if (scene == 6) {
      // SCENE 6
      LIGHT_X = 0;
      LIGHT_Y = -10000;
      LIGHT_Z = 10000;
      AffineTransform translation
          = AffineTransform.getTranslationTransform(2300, -1500, 8700);
      AffineTransform rotationY
          = AffineTransform.getRotateAroundYTransform(0.5);
      AffineTransform rotationX
          = AffineTransform.getRotateAroundXTransform(-0.2);
      transform = rotationY.multiply(rotationX);
      transform = translation.multiply(transform);
    }

    for(int i = 0; i < camera.length; i++) {
      camera[i] = transform.multiply(camera[i]);
    }

    Vector leftVector = camera[2].subtract(camera[0]);
    Vector rightVector = camera[3].subtract(camera[1]);

    Solid wheel = new Solid();
    wheel.translate(0, -WHEEL_RADIUS, 0);
    wheel.isSphere = false;
    wheel.radius = WHEEL_RADIUS;
    wheel.length = WHEEL_RADIUS * 0.1;
    wheel.ambientRed = 0x35;
    wheel.ambientGreen = 0x31;
    wheel.ambientBlue = 0x2E;
    wheel.diffuseRed = 0x89;
    wheel.diffuseGreen = 0x8D;
    wheel.diffuseBlue = 0x98;

    Solid cylinder1 = new Solid();
    cylinder1.isSphere = false;
    cylinder1.rotateAroundY(Math.PI / 2.0);
    cylinder1.radius = WHEEL_RADIUS;
    cylinder1.length = WHEEL_RADIUS * 0.1;

    Solid rightFoot = new Solid()
        .scale(WHEEL_RADIUS / 4, WHEEL_RADIUS * 0.05, WHEEL_RADIUS * 0.6)
        .translate(-WHEEL_RADIUS * 0.5, 0, -WHEEL_RADIUS * 0.2);
    Solid rightAnkle = new Solid()
        .scale(WHEEL_RADIUS * 0.15, WHEEL_RADIUS * 0.1, WHEEL_RADIUS * 0.15)
        .translate(-WHEEL_RADIUS * 0.5, -WHEEL_RADIUS * 0.1, 0);
    Solid rightLowerLeg = new Solid()
        .scale(WHEEL_RADIUS * 0.3, WHEEL_RADIUS, WHEEL_RADIUS * 0.3)
        .translate(-WHEEL_RADIUS * 0.5, -WHEEL_RADIUS, 0);
    Solid rightKnee = new Solid()
        .scale(WHEEL_RADIUS * 0.3, WHEEL_RADIUS * 0.3, WHEEL_RADIUS * 0.3)
        .translate(-WHEEL_RADIUS * 0.5, -WHEEL_RADIUS * 2, 0);
    Solid rightUpperLeg = new Solid()
        .scale(WHEEL_RADIUS * 0.35, WHEEL_RADIUS * 1.2, WHEEL_RADIUS * 0.35)
        .translate(-WHEEL_RADIUS * 0.5, -WHEEL_RADIUS * 1.2 - WHEEL_RADIUS * 2, 0);
    Solid rightHip = new Solid()
        .scale(WHEEL_RADIUS * 0.35, WHEEL_RADIUS * 0.35, WHEEL_RADIUS * 0.35)
        .translate(-WHEEL_RADIUS * 0.5, -WHEEL_RADIUS * 1.2 * 2 - WHEEL_RADIUS * 2, 0);

    Solid leftFoot = new Solid()
        .scale(WHEEL_RADIUS / 4, WHEEL_RADIUS * 0.05, WHEEL_RADIUS * 0.6)
        .translate(WHEEL_RADIUS * 0.5, 0, -WHEEL_RADIUS * 0.2);
    Solid leftAnkle = new Solid()
        .scale(WHEEL_RADIUS * 0.15, WHEEL_RADIUS * 0.1, WHEEL_RADIUS * 0.15)
        .translate(WHEEL_RADIUS * 0.5, -WHEEL_RADIUS * 0.1, 0);
    Solid leftLowerLeg = new Solid()
        .scale(WHEEL_RADIUS * 0.3, WHEEL_RADIUS, WHEEL_RADIUS * 0.3)
        .translate(WHEEL_RADIUS * 0.5, -WHEEL_RADIUS, 0);
    Solid leftKnee = new Solid()
        .scale(WHEEL_RADIUS * 0.3, WHEEL_RADIUS * 0.3, WHEEL_RADIUS * 0.3)
        .translate(WHEEL_RADIUS * 0.5, -WHEEL_RADIUS * 2, 0);
    Solid leftUpperLeg = new Solid()
        .scale(WHEEL_RADIUS * 0.35, WHEEL_RADIUS * 1.2, WHEEL_RADIUS * 0.35)
        .translate(WHEEL_RADIUS * 0.5, -WHEEL_RADIUS * 1.2 - WHEEL_RADIUS * 2, 0);
    Solid leftHip = new Solid()
        .scale(WHEEL_RADIUS * 0.35, WHEEL_RADIUS * 0.35, WHEEL_RADIUS * 0.35)
        .translate(WHEEL_RADIUS * 0.5, -WHEEL_RADIUS * 1.2 * 2 - WHEEL_RADIUS * 2, 0);

    Solid groin = new Solid()
        .scale(WHEEL_RADIUS * 0.5, WHEEL_RADIUS * 0.35, WHEEL_RADIUS * 0.35)
        .translate(0, -WHEEL_RADIUS * 1.2 * 2 - WHEEL_RADIUS * 2, 0);
    groin.isSphere = false;
    Solid ab1 = new Solid()
        .scale(WHEEL_RADIUS * 0.6, WHEEL_RADIUS * 0.2, WHEEL_RADIUS * 0.35)
        .translate(0, -WHEEL_RADIUS * 4.8, 0);
    Solid ab2 = new Solid()
        .scale(WHEEL_RADIUS * 0.6, WHEEL_RADIUS * 0.2, WHEEL_RADIUS * 0.35)
        .translate(0, -WHEEL_RADIUS * (4.8 + 0.15), 0);
    Solid ab3 = new Solid()
        .scale(WHEEL_RADIUS * 0.6, WHEEL_RADIUS * 0.2, WHEEL_RADIUS * 0.35)
        .translate(0, -WHEEL_RADIUS * (4.8 + 0.3), 0);
    Solid ab4 = new Solid()
        .scale(WHEEL_RADIUS * 0.6, WHEEL_RADIUS * 0.2, WHEEL_RADIUS * 0.35)
        .translate(0, -WHEEL_RADIUS * (4.8 + 0.45), 0);
    Solid torso = new Solid()
        .scale(WHEEL_RADIUS * 0.7, WHEEL_RADIUS, WHEEL_RADIUS * 0.35)
        .translate(0, -WHEEL_RADIUS * (5.25 + 0.6), 0);
    Solid chest = new Solid()
        .scale(WHEEL_RADIUS, WHEEL_RADIUS * 0.35, WHEEL_RADIUS * 0.35)
        .translate(0, -WHEEL_RADIUS * 6.65, 0);
    chest.isSphere = false;
    Solid neck = new Solid()
        .scale(WHEEL_RADIUS * 0.3, WHEEL_RADIUS * 0.3, WHEEL_RADIUS * 0.3)
        .rotateAroundZ(Math.PI / 2.0)
        .translate(0, -WHEEL_RADIUS * 7.0, 0);
    neck.isSphere = false;
    Solid head = new Solid()
        .scale(WHEEL_RADIUS * 0.5, WHEEL_RADIUS * 0.7, WHEEL_RADIUS * 0.5)
        .rotateAroundX(-0.35)
        .translate(0, -WHEEL_RADIUS * 7.75, 0);

    Solid rightShoulder = new Solid()
        .scale(WHEEL_RADIUS * 0.35, WHEEL_RADIUS * 0.35, WHEEL_RADIUS * 0.35)
        .translate(-WHEEL_RADIUS, -WHEEL_RADIUS * 6.65, 0);
    Solid rightUpperArm = new Solid()
        .scale(WHEEL_RADIUS * 0.25, WHEEL_RADIUS * 0.75, WHEEL_RADIUS * 0.25)
        .translate(-WHEEL_RADIUS * 1.1, -WHEEL_RADIUS * (6.65 - 0.75), 0);
    Solid rightElbo = new Solid()
        .scale(WHEEL_RADIUS * 0.25, WHEEL_RADIUS * 0.25, WHEEL_RADIUS * 0.25)
        .translate(-WHEEL_RADIUS * 1.1, -WHEEL_RADIUS * 5.15, 0);
    Solid rightLowerArm = new Solid()
        .scale(WHEEL_RADIUS * 0.2, WHEEL_RADIUS * 0.75, WHEEL_RADIUS * 0.2)
        .translate(-WHEEL_RADIUS * 1.1, -WHEEL_RADIUS * 4.4, 0);
    Solid rightWrist = new Solid()
        .scale(WHEEL_RADIUS * 0.2, WHEEL_RADIUS * 0.2, WHEEL_RADIUS * 0.2)
        .translate(-WHEEL_RADIUS * 1.1, -WHEEL_RADIUS * 3.65, 0);
    Solid rightHand = new Solid()
        .scale(WHEEL_RADIUS * 0.1, WHEEL_RADIUS * 0.42, WHEEL_RADIUS * 0.2)
        .translate(-WHEEL_RADIUS * 1.1, -WHEEL_RADIUS * 3.35, 0);

    Solid leftShoulder = new Solid()
        .scale(WHEEL_RADIUS * 0.35, WHEEL_RADIUS * 0.35, WHEEL_RADIUS * 0.35)
        .translate(WHEEL_RADIUS, -WHEEL_RADIUS * 6.65, 0);
    Solid leftUpperArm = new Solid()
        .scale(WHEEL_RADIUS * 0.25, WHEEL_RADIUS * 0.75, WHEEL_RADIUS * 0.25)
        .translate(WHEEL_RADIUS * 1.1, -WHEEL_RADIUS * (6.65 - 0.75), 0);
    Solid leftElbo = new Solid()
        .scale(WHEEL_RADIUS * 0.25, WHEEL_RADIUS * 0.25, WHEEL_RADIUS * 0.25)
        .translate(WHEEL_RADIUS * 1.1, -WHEEL_RADIUS * 5.15, 0);
    Solid leftLowerArm = new Solid()
        .scale(WHEEL_RADIUS * 0.2, WHEEL_RADIUS * 0.75, WHEEL_RADIUS * 0.2)
        .translate(WHEEL_RADIUS * 1.1, -WHEEL_RADIUS * 4.4, 0);
    Solid leftWrist = new Solid()
        .scale(WHEEL_RADIUS * 0.2, WHEEL_RADIUS * 0.2, WHEEL_RADIUS * 0.2)
        .translate(WHEEL_RADIUS * 1.1, -WHEEL_RADIUS * 3.65, 0);
    Solid leftHand = new Solid()
        .scale(WHEEL_RADIUS * 0.1, WHEEL_RADIUS * 0.42, WHEEL_RADIUS * 0.2)
        .translate(WHEEL_RADIUS * 1.1, -WHEEL_RADIUS * 3.35, 0);

    Solid[] solids = { wheel, rightFoot, rightAnkle, rightLowerLeg, rightKnee, 
        rightUpperLeg, rightHip, leftFoot, leftAnkle, leftLowerLeg, leftKnee,
        leftUpperLeg, leftHip, groin, ab1, ab2, ab3, ab4, torso, chest,
        neck, head, rightShoulder, rightUpperArm, rightElbo, rightLowerArm,
        rightWrist, rightHand, leftShoulder, leftUpperArm, leftElbo,
        leftLowerArm, leftWrist, leftHand };

    Solid[] rightArmSolids = { rightUpperArm, rightElbo, rightLowerArm,
        rightWrist, rightHand };
    Solid[] rightLowerArmSolids = { rightLowerArm, rightWrist, rightHand };
    Solid[] leftArmSolids = { leftUpperArm, leftElbo, leftLowerArm,
        leftWrist, leftHand };
    Solid[] leftLowerArmSolids = { leftLowerArm, leftWrist, leftHand };

    Solid[] rightLegSolids = { rightFoot, rightAnkle, rightLowerLeg, rightKnee,
        rightUpperLeg };
    Solid[] rightLowerLegSolids = { rightFoot, rightAnkle, rightLowerLeg };
    Solid[] leftLegSolids = { leftFoot, leftAnkle, leftLowerLeg, leftKnee,
        leftUpperLeg };
    Solid[] leftLowerLegSolids = { leftFoot, leftAnkle, leftLowerLeg };
    
    double rightFramePercent = frame / 60.0;
    double rightUpperArmAngle = -(Math.PI / 8) * Math.sin(Math.PI * rightFramePercent);
    double rightArmAngle = (Math.PI / 8) * Math.cos(2.0 * Math.PI * rightFramePercent);
    double leftFramePercent = ((30 + frame) % 60) / 60.0;
    double leftUpperArmAngle = -(Math.PI / 8) * Math.sin(Math.PI * leftFramePercent);
    double leftArmAngle = (Math.PI / 8) * Math.cos(2.0 * Math.PI * leftFramePercent);

    double footY = -WHEEL_RADIUS + 0.65 * WHEEL_RADIUS * Math.cos(3.0 * Math.PI / 2.0 + 2.0 * Math.PI * rightFramePercent);
    double footZ = -0.65 * WHEEL_RADIUS * Math.sin(3.0 * Math.PI / 2.0 + 2.0 * Math.PI * rightFramePercent);
    double footDist = Math.hypot(footY + WHEEL_RADIUS * 4.4, footZ);
    double footAngle1 = Math.atan2(footZ, footY + WHEEL_RADIUS * 4.4);
    
    double footDx = (footDist * footDist + (WHEEL_RADIUS * 2.4) * (WHEEL_RADIUS * 2.4)
        - (WHEEL_RADIUS * 2.0) * (WHEEL_RADIUS * 2.0)) / (2.0 * footDist);
    double footDy = Math.sqrt((WHEEL_RADIUS * 2.4) * (WHEEL_RADIUS * 2.4) - footDx * footDx);
    double footAngle2 = Math.atan2(footDy, footDx);
    double footAngle3 = Math.PI - ((Math.PI * 0.5 - Math.atan2(footDy, footDist - footDx))
        + (Math.PI * 0.5 - footAngle2));

    double footY2 = -WHEEL_RADIUS + 0.65 * WHEEL_RADIUS * Math.cos(Math.PI / 2.0 + 2.0 * Math.PI * rightFramePercent);
    double footZ2 = -0.65 * WHEEL_RADIUS * Math.sin(Math.PI / 2.0 + 2.0 * Math.PI * rightFramePercent);
    double footDist2 = Math.hypot(footY2 + WHEEL_RADIUS * 4.4, footZ2);
    double footAngle12 = Math.atan2(footZ2, footY2 + WHEEL_RADIUS * 4.4);

    double footDx2 = (footDist2 * footDist2 + (WHEEL_RADIUS * 2.4) * (WHEEL_RADIUS * 2.4)
        - (WHEEL_RADIUS * 2.0) * (WHEEL_RADIUS * 2.0)) / (2.0 * footDist2);
    double footDy2 = Math.sqrt((WHEEL_RADIUS * 2.4) * (WHEEL_RADIUS * 2.4) - footDx2 * footDx2);
    double footAngle22 = Math.atan2(footDy2, footDx2);
    double footAngle32 = Math.PI - ((Math.PI * 0.5 - Math.atan2(footDy2, footDist2 - footDx2))
        + (Math.PI * 0.5 - footAngle22));

    for(Solid solid : rightLowerArmSolids) {
      solid.rotateAroundX(rightUpperArmAngle, -WHEEL_RADIUS * 5.15, 0);
    }
    for(Solid solid : rightArmSolids) {
      solid.rotateAroundX(rightArmAngle, -WHEEL_RADIUS * 6.65, 0);
    }
    for(Solid solid : leftLowerArmSolids) {
      solid.rotateAroundX(leftUpperArmAngle, -WHEEL_RADIUS * 5.15, 0);
    }
    for(Solid solid : leftArmSolids) {
      solid.rotateAroundX(leftArmAngle, -WHEEL_RADIUS * 6.65, 0);
    }

    for(Solid solid : rightLowerLegSolids) {
      solid.rotateAroundX(footAngle3, -WHEEL_RADIUS * 2, 0);
    }
    for(Solid solid : rightLegSolids) {
      solid.rotateAroundX(-(footAngle1 + footAngle2), -WHEEL_RADIUS * 4.4, 0);
    }
    for(Solid solid : leftLowerLegSolids) {
      solid.rotateAroundX(footAngle32, -WHEEL_RADIUS * 2, 0);
    }
    for(Solid solid : leftLegSolids) {
      solid.rotateAroundX(-(footAngle12 + footAngle22), -WHEEL_RADIUS * 4.4, 0);
    }

    double percent = frame / 60.0;
    double reversePercent = 1 - percent;
    TransformPair[] transformPairs = new TransformPair[64];
    for(int i = 0; i < transformPairs.length; i++) {
      transformPairs[i] = new TransformPair();
    }
    transformPairs[HilbertCurve.DOWN].translate(SQUARE_SIZE * 0.5, 0, SQUARE_SIZE * reversePercent);
    transformPairs[HilbertCurve.UP].rotateAroundY(Math.PI).translate(SQUARE_SIZE * 0.5, 0, SQUARE_SIZE * percent);
    transformPairs[HilbertCurve.LEFT].rotateAroundY(-Math.PI * 0.5).translate(SQUARE_SIZE * percent, 0, SQUARE_SIZE * 0.5);
    transformPairs[HilbertCurve.RIGHT].rotateAroundY(Math.PI * 0.5).translate(SQUARE_SIZE * reversePercent, 0, SQUARE_SIZE * 0.5);
    transformPairs[(HilbertCurve.DOWN << 3) | HilbertCurve.DOWN] = transformPairs[HilbertCurve.DOWN];
    transformPairs[(HilbertCurve.UP << 3) | HilbertCurve.UP] = transformPairs[HilbertCurve.UP];
    transformPairs[(HilbertCurve.LEFT << 3) | HilbertCurve.LEFT] = transformPairs[HilbertCurve.LEFT];
    transformPairs[(HilbertCurve.RIGHT << 3) | HilbertCurve.RIGHT] = transformPairs[HilbertCurve.RIGHT];
    
    transformPairs[(HilbertCurve.RIGHT << 3) | HilbertCurve.UP]
        .rotateAroundZ(Math.sin(Math.PI * percent) * Math.PI / 16.0)
        .rotateAroundY(Math.PI - percent * Math.PI * 0.5)
        .translate(0.5 * SQUARE_SIZE * Math.cos(0.5 * Math.PI * percent), 0, 0.5 * SQUARE_SIZE * Math.sin(0.5 * Math.PI * percent));
    transformPairs[(HilbertCurve.DOWN << 3) | HilbertCurve.LEFT]
        .rotateAroundZ(-Math.sin(Math.PI * percent) * Math.PI / 16.0)
        .rotateAroundY(-reversePercent * 0.5 * Math.PI)
        .translate(0.5 * SQUARE_SIZE * Math.cos(0.5 * Math.PI * reversePercent), 0, 0.5 * SQUARE_SIZE * Math.sin(0.5 * Math.PI * reversePercent));

    transformPairs[(HilbertCurve.UP << 3) | HilbertCurve.LEFT]
        .rotateAroundZ(Math.sin(Math.PI * percent) * Math.PI / 16.0)
        .rotateAroundY(3.0 * Math.PI / 2.0 - percent * Math.PI * 0.5)
        .translate(0.5 * SQUARE_SIZE * Math.cos(3.0 * Math.PI / 2.0 + 0.5 * Math.PI * percent), 0, SQUARE_SIZE + 0.5 * SQUARE_SIZE * Math.sin(3.0 * Math.PI / 2.0 + 0.5 * Math.PI * percent));
    transformPairs[(HilbertCurve.RIGHT << 3) | HilbertCurve.DOWN]
        .rotateAroundZ(-Math.sin(Math.PI * percent) * Math.PI / 16.0)
        .rotateAroundY(0.5 * Math.PI * percent)
        .translate(0.5 * SQUARE_SIZE * Math.cos(3.0 * Math.PI / 2.0 + 0.5 * Math.PI * reversePercent), 0, SQUARE_SIZE + 0.5 * SQUARE_SIZE * Math.sin(3.0 * Math.PI / 2.0 + 0.5 * Math.PI * reversePercent));

    transformPairs[(HilbertCurve.LEFT << 3) | HilbertCurve.DOWN]
        .rotateAroundZ(Math.sin(Math.PI * percent) * Math.PI / 16.0)
        .rotateAroundY(-percent * Math.PI * 0.5)
        .translate(SQUARE_SIZE + 0.5 * SQUARE_SIZE * Math.cos(Math.PI + 0.5 * Math.PI * percent), 0, SQUARE_SIZE + 0.5 * SQUARE_SIZE * Math.sin(Math.PI + 0.5 * Math.PI * percent));
    transformPairs[(HilbertCurve.UP << 3) | HilbertCurve.RIGHT]
        .rotateAroundZ(-Math.sin(Math.PI * percent) * Math.PI / 16.0)
        .rotateAroundY(Math.PI * 0.5 + percent * Math.PI * 0.5)
        .translate(SQUARE_SIZE + 0.5 * SQUARE_SIZE * Math.cos(Math.PI + 0.5 * Math.PI * reversePercent), 0, SQUARE_SIZE + 0.5 * SQUARE_SIZE * Math.sin(Math.PI + 0.5 * Math.PI * reversePercent));

    transformPairs[(HilbertCurve.DOWN << 3) | HilbertCurve.RIGHT]
        .rotateAroundZ(Math.sin(Math.PI * percent) * Math.PI / 16.0)
        .rotateAroundY(Math.PI * 0.5 - percent * Math.PI * 0.5)
        .translate(SQUARE_SIZE + 0.5 * SQUARE_SIZE * Math.cos(Math.PI * 0.5 + 0.5 * Math.PI * percent), 0, 0.5 * SQUARE_SIZE * Math.sin(Math.PI * 0.5 + 0.5 * Math.PI * percent));
    transformPairs[(HilbertCurve.LEFT << 3) | HilbertCurve.UP]
        .rotateAroundZ(-Math.sin(Math.PI * percent) * Math.PI / 16.0)
        .rotateAroundY(Math.PI + percent * Math.PI * 0.5)
        .translate(SQUARE_SIZE + 0.5 * SQUARE_SIZE * Math.cos(Math.PI * 0.5 + 0.5 * Math.PI * reversePercent), 0, 0.5 * SQUARE_SIZE * Math.sin(Math.PI * 0.5 + 0.5 * Math.PI * reversePercent));


    int[][] visitOrder = { 
      { 0, 0 }, { -1, 0 }, { 0, -1 }, { 1, 0 }, { 0, 1 },
      { 1, 1 }, { -1, -1 }, { 1, -1 }, { -1, 1 } };
    int[][] visitedIndexes = new int[1024][2];
    int visitedCount = 0;

    TransformPair transformPair = null;
    SquareIntersect squareIntersect = new SquareIntersect();
    SolidIntersect solidIntersect = new SolidIntersect();
    Vector bestPoint = new Vector();
    Vector bestNormal = new Vector();
    Vector light = new Vector();
    Vector reflected = new Vector();
    Vector view = new Vector();
    Vector shadowOrigin = new Vector();
    Vector shadowRay = new Vector();
    Solid bestSolid = null;

    for (int y = 0; y < HEIGHT; y++) {
      System.out.println("rendering: scene = " + scene + ", frame = " + frame + ", row = " + y);
      for (int ys = 0; ys < SAMPLES; ys++) {
        double verticlePercent = ( ((double)ys) / SAMPLES
            + ((double) y)) / HEIGHT;
        Vector leftPoint = camera[0].add(leftVector.multiply(verticlePercent));
        Vector rightPoint = camera[1].add(rightVector.multiply(verticlePercent));
        Vector acrossVector = rightPoint.subtract(leftPoint);
        for (int x = 0; x < WIDTH; x++) {
          outter: for (int xs = 0; xs < SAMPLES; xs++) {

            for(int i = 0; i < visitedCount; i++) {
              visited[visitedIndexes[i][1]][visitedIndexes[i][0]] = false;
            }
            visitedCount = 0;

            double horizontalPercent = (((double)xs) / SAMPLES
                + ((double)x)) / WIDTH;
            Vector point = leftPoint.add(acrossVector.multiply(horizontalPercent));
            Vector eye = camera[4];
            Vector ray = point.subtract(eye);

            boolean intersected = false;

            if (renderMen) {

            squareIntersect.vx = ray.getX();
            squareIntersect.vy = ray.getY();
            squareIntersect.vz = ray.getZ();

            if (eye.getY() >= RIDER_Y) {
              squareIntersect.curveX = (int) Math.floor(eye.getX() / SQUARE_SIZE);
              squareIntersect.curveZ = (int) Math.floor(eye.getZ() / SQUARE_SIZE);
              squareIntersect.x = eye.getX();
              squareIntersect.y = eye.getY();
              squareIntersect.z = eye.getZ();
            } else {
              double t = (RIDER_Y - eye.getY()) / ray.getY();
              if (t > 0) {
                double surfaceX = eye.getX() + t * ray.getX();
                double surfaceZ = eye.getZ() + t * ray.getZ();
                squareIntersect.curveX = (int) Math.floor(surfaceX / SQUARE_SIZE);
                squareIntersect.curveZ = (int) Math.floor(surfaceZ / SQUARE_SIZE);
                squareIntersect.x = surfaceX;
                squareIntersect.y = RIDER_Y;
                squareIntersect.z = surfaceZ;
              } else {
                squareIntersect.y = RIDER_Y * 2;
              }
            }

            while(!intersected
                && squareIntersect.curveX > -10
                && squareIntersect.curveX < 776
                && squareIntersect.curveZ > -10
                && squareIntersect.curveZ < 776
                && squareIntersect.y < 0
                && squareIntersect.y >= RIDER_Y
                && (squareIntersect.vz != 0 || squareIntersect.vx != 0)) {

              double bestDistance2 = Double.MAX_VALUE;
              bestSolid = null;
              int bestVisitX = 0;
              int bestVisitZ = 0;

              for(int i = 0; i < visitOrder.length; i++) {
                int visitX = squareIntersect.curveX + visitOrder[i][0];
                int visitZ = squareIntersect.curveZ + visitOrder[i][1];
                if (visitX < 0 || visitZ < 0 || visitX >= visited.length
                    || visitZ >= visited.length || visited[visitZ][visitX]
                    || hilbertCurve[visitZ][visitX] == 0) {
                  continue;
                }
                visitedIndexes[visitedCount][0] = visitX;
                visitedIndexes[visitedCount][1] = visitZ;
                visitedCount++;
                visited[visitZ][visitX] = true;

                transformPair = transformPairs[hilbertCurve[visitZ][visitX]];
                transformPair.offset(visitX * SQUARE_SIZE, 0, visitZ * SQUARE_SIZE);

                for(Solid solid : solids) {
                  findSolidIntersect(eye, ray, solid, solidIntersect, transformPair, true);
                  if (solidIntersect.intersected) {
                    double distance2 = Vector.distance2(eye, solidIntersect.point);
                    if (distance2 < bestDistance2) {
                      bestDistance2 = distance2;
                      bestPoint.copyFrom(solidIntersect.point);
                      bestNormal.copyFrom(solidIntersect.normal);
                      bestSolid = solid;
                      bestVisitX = visitX;
                      bestVisitZ = visitZ;
                    }
                  }
                }

                transformPair.undoOffset();
              }

              if (bestSolid != null) {
                intersected = true;
                light.setX(LIGHT_X);
                light.setY(LIGHT_Y);
                light.setZ(LIGHT_Z);
                light.subtractVector(bestPoint);
                light.normalize();

                double dotLN = light.dot(bestNormal);
                if (dotLN > 0) {
                  transformPair = transformPairs[hilbertCurve[bestVisitZ][bestVisitX]];
                  transformPair.offset(bestVisitX * SQUARE_SIZE, 0, bestVisitZ * SQUARE_SIZE);
                  for(Solid solid : solids) {
                    if (solid != bestSolid) {
                      findSolidIntersect(bestPoint, light, solid, solidIntersect, transformPair, false);
                      if (solidIntersect.intersected) {
                        pixels[y][x][0] += bestSolid.ambientRed;
                        pixels[y][x][1] += bestSolid.ambientGreen;
                        pixels[y][x][2] += bestSolid.ambientBlue;
                        transformPair.undoOffset();
                        continue outter;
                      }
                    }
                  }
                  transformPair.undoOffset();

                  reflected.copyFrom(bestNormal);
                  reflected.scale(2.0 * dotLN);
                  reflected.subtractVector(light);
                  reflected.normalize();

                  view.copyFrom(camera[4]);
                  view.subtractVector(bestPoint);
                  view.normalize();

                  double specular = Math.pow(Math.max(0, reflected.dot(view)), 75);

                  pixels[y][x][0] += Math.min(255,
                      bestSolid.ambientRed + dotLN * bestSolid.diffuseRed + specular * 255);
                  pixels[y][x][1] += Math.min(255,
                      bestSolid.ambientGreen + dotLN * bestSolid.diffuseGreen + specular * 255);
                  pixels[y][x][2] += Math.min(255,
                      bestSolid.ambientBlue + dotLN * bestSolid.diffuseBlue + specular * 255);
                } else {
                  pixels[y][x][0] += bestSolid.ambientRed;
                  pixels[y][x][1] += bestSolid.ambientGreen;
                  pixels[y][x][2] += bestSolid.ambientBlue;
                }
              }

              findSquareIntersect(squareIntersect);
            }
            }

            if (!intersected) {
              double t = -eye.getY() / ray.getY();
              if (t > 0) {
                double surfaceX = eye.getX() + t * ray.getX();
                double surfaceZ = eye.getZ() + t * ray.getZ();
                int curveX = (int) Math.floor(surfaceX / SQUARE_SIZE);
                int curveZ = (int) Math.floor(surfaceZ / SQUARE_SIZE);
                shadowOrigin.setX(surfaceX);
                shadowOrigin.setY(0);
                shadowOrigin.setZ(surfaceZ);
                shadowRay.setX(LIGHT_X - surfaceX);
                shadowRay.setY(LIGHT_Y);
                shadowRay.setZ(LIGHT_Z - surfaceZ);

                if (curveX >= 0 && curveX < 766 && curveZ >= 0 && curveZ < 766
                        && hilbertCurve[curveZ][curveX] > 0) {

                  intersected = true;

                  for(int i = 0; i < visitOrder.length; i++) {
                    int visitX = curveX + visitOrder[i][0];
                    int visitZ = curveZ + visitOrder[i][1];
                    if (visitX < 0 || visitZ < 0 || visitX >= visited.length
                        || visitZ >= visited.length
                        || hilbertCurve[visitZ][visitX] == 0) {
                      continue;
                    }

                    transformPair = transformPairs[hilbertCurve[visitZ][visitX]];
                    transformPair.offset(visitX * SQUARE_SIZE, 0, visitZ * SQUARE_SIZE);
                    for(Solid solid : solids) {
                      findSolidIntersect(shadowOrigin, shadowRay, solid,
                          solidIntersect, transformPair, false);
                      if (solidIntersect.intersected) {
                        transformPair.undoOffset();
                        pixels[y][x][0] += 0x45;
                        pixels[y][x][1] += 0x45;
                        pixels[y][x][2] += 0x45;
                        continue outter;
                      }
                    }
                    transformPair.undoOffset();
                  }

                  pixels[y][x][0] += 0xFD;
                  pixels[y][x][1] += 0xFD;
                  pixels[y][x][2] += 0xFD;
                  continue outter;
                }

                if (curveX > -10 && curveX < 776 && curveZ > -10 && curveZ < 776) {

                  squareIntersect.x = surfaceX;
                  squareIntersect.y = 0;
                  squareIntersect.z = surfaceZ;
                  squareIntersect.vx = ray.getX();
                  squareIntersect.vy = ray.getY();
                  squareIntersect.vz = ray.getZ();
                  squareIntersect.curveX = curveX;
                  squareIntersect.curveZ = curveZ;

                  while (squareIntersect.curveX > -10
                      && squareIntersect.curveX < 776
                      && squareIntersect.curveZ > -10
                      && squareIntersect.curveZ < 776
                      && squareIntersect.y < GROUND_Y) {
                    findSquareIntersect(squareIntersect);
                    if (squareIntersect.curveX >= 0
                        && squareIntersect.curveX < 766
                        && squareIntersect.curveZ >= 0
                        && squareIntersect.curveZ < 766
                        && hilbertCurve[squareIntersect.curveZ][squareIntersect.curveX] > 0
                        && squareIntersect.y < GROUND_Y) {

                      intersected = true;
                      double wallRed = 0x46;
                      double wallGreen = 0x46;
                      double wallBlue = 0x46;
                      double wallX = squareIntersect.x;
                      double wallY = squareIntersect.y;
                      double wallZ = squareIntersect.z;
                      double wallNormalX = squareIntersect.normalX;
                      double wallNormalY = 0;
                      double wallNormalZ = squareIntersect.normalZ;
                      double lightX = LIGHT_X - wallX;
                      double lightY = LIGHT_Y - wallY;
                      double lightZ = LIGHT_Z - wallZ;
                      double lightMag = magnitude(lightX, lightY, lightZ);
                      lightX /= lightMag;
                      lightY /= lightMag;
                      lightZ /= lightMag;

                      double dot = dot(wallNormalX, wallNormalY, wallNormalZ,
                              lightX, lightY, lightZ);
                      if (dot > 0) {
                        squareIntersect.vx = lightX;
                        squareIntersect.vy = lightY;
                        squareIntersect.vz = lightZ;
                        squareIntersect.curveX = squareIntersect.lastCurveX;
                        squareIntersect.curveZ = squareIntersect.lastCurveZ;
                        while (squareIntersect.curveX > -10
                            && squareIntersect.curveX < 776
                            && squareIntersect.curveZ > -10
                            && squareIntersect.curveZ < 776
                            && squareIntersect.y > 0) {
                          findSquareIntersect(squareIntersect);
                          if (squareIntersect.curveX >= 0
                              && squareIntersect.curveX < 766
                              && squareIntersect.curveZ >= 0
                              && squareIntersect.curveZ < 766
                              && hilbertCurve[squareIntersect.curveZ][squareIntersect.curveX] > 0) {
                            break;
                          }
                        }

                        if (squareIntersect.y < 0) {
                          wallBlue = wallGreen = wallRed
                              = Math.min(255, 0x58 + (0xE9 - 0x58) * dot);
                        }
                      }

                      pixels[y][x][0] += wallRed;
                      pixels[y][x][1] += wallGreen;
                      pixels[y][x][2] += wallBlue;

                      continue outter;
                    }
                  }
                }
              }
            }

            if (!intersected) {
              double t = (SKY_Y - eye.getY()) / ray.getY();
              if (t > 0) {
                intersected = true;
                double surfaceX = eye.getX() + t * ray.getX();
                double surfaceZ = eye.getZ() + t * ray.getZ();
                getPixel(surfaceX * 0.01, surfaceZ * 0.01, skyMap, mapPixel, tempMapPixels);
                pixels[y][x][0] += mapPixel[0];
                pixels[y][x][1] += mapPixel[1];
                pixels[y][x][2] += mapPixel[2];
                continue outter;
              }
            }

            if (!intersected) {
              double t = (GROUND_Y - eye.getY()) / ray.getY();
              if (t > 0) {
                intersected = true;
                double surfaceX = eye.getX() + t * ray.getX();
                double surfaceZ = eye.getZ() + t * ray.getZ();
                if (surfaceX > 700 * SQUARE_SIZE || surfaceZ > 700 * SQUARE_SIZE) {
                  getPixel(surfaceX * 0.01, surfaceZ * 0.01, skyMap, mapPixel, tempMapPixels);
                } else {
                  getPixel(surfaceX * 0.01, surfaceZ * 0.01, earthMap, mapPixel, tempMapPixels);
                }
                pixels[y][x][0] += mapPixel[0];
                pixels[y][x][1] += mapPixel[1];
                pixels[y][x][2] += mapPixel[2];
                continue outter;
              }
            }
          }
        }
      }      
    }

    saveImage(path + File.separator + "render" + scene + "_" + frame + ".png",
        pixels);
  }

  public void render2() throws Throwable  {

    byte[][] hilbertCurve = HilbertCurve.generate(8);

    double[][][] pixels = new double[HEIGHT][WIDTH][3];

    Vector upperLeft = new Vector(-320, -180, 0);
    Vector upperRight = new Vector(320, -180, 0);
    Vector lowerLeft = new Vector(-320, 180, 0);
    Vector lowerRight = new Vector(320, 180, 0);
    Vector viewer = new Vector(0, 0, -VIEWER);
    Vector[] camera = { upperLeft, upperRight, lowerLeft, lowerRight, viewer };

    AffineTransform translation
        = AffineTransform.getTranslationTransform(2300, -1500, 8700);
    AffineTransform rotationY
        = AffineTransform.getRotateAroundYTransform(0.5);
    AffineTransform rotationX
        = AffineTransform.getRotateAroundXTransform(-0.2);
    AffineTransform transform = rotationY.multiply(rotationX);
    transform = translation.multiply(transform);
    for(int i = 0; i < camera.length; i++) {
      camera[i] = transform.multiply(camera[i]);
    }

    Vector leftVector = camera[2].subtract(camera[0]);
    Vector rightVector = camera[3].subtract(camera[1]);

    SquareIntersect squareIntersect = new SquareIntersect();
    SphereIntersect sphereIntersect = new SphereIntersect();

    for (int y = 0; y < HEIGHT; y++) {
      for (int ys = 0; ys < SAMPLES; ys++) {
        double verticlePercent = ( ((double)ys) / SAMPLES
            + ((double) y)) / HEIGHT;
        Vector leftPoint = camera[0].add(leftVector.multiply(verticlePercent));
        Vector rightPoint = camera[1].add(rightVector.multiply(verticlePercent));
        Vector acrossVector = rightPoint.subtract(leftPoint);
        for (int x = 0; x < WIDTH; x++) {
          outter: for (int xs = 0; xs < SAMPLES; xs++) {

            double horizontalPercent = (((double)xs) / SAMPLES
                + ((double)x)) / WIDTH;
            Vector point = leftPoint.add(acrossVector.multiply(horizontalPercent));
            Vector eye = camera[4];
            Vector ray = point.subtract(eye);

            double t = (RIDER_Y - eye.getY()) / ray.getY();

            /*if (t > 0) {
              double surfaceX = eye.getX() + t * ray.getX();
              double surfaceY = eye.getY() + t * ray.getY();
              double surfaceZ = eye.getZ() + t * ray.getZ();
              int curveX = (int) Math.floor(surfaceX / SQUARE_SIZE);
              int curveZ = (int) Math.floor(surfaceZ / SQUARE_SIZE);
              
              squareIntersect.x = surfaceX;
              squareIntersect.y = surfaceY;
              squareIntersect.z = surfaceZ;
              squareIntersect.vx = ray.getX();
              squareIntersect.vy = ray.getY();
              squareIntersect.vz = ray.getZ();
              squareIntersect.curveX = curveX;
              squareIntersect.curveZ = curveZ;

              while (squareIntersect.curveX > -10
                  && squareIntersect.curveX < 776
                  && squareIntersect.curveZ > -10
                  && squareIntersect.curveZ < 776
                  && squareIntersect.y < 0) {
                if (squareIntersect.curveX >= 0
                    && squareIntersect.curveX < 766
                    && squareIntersect.curveZ >= 0
                    && squareIntersect.curveZ < 766
                    && hilbertCurve[squareIntersect.curveZ][squareIntersect.curveX] > 0) {
                  findSphereIntersect(
                      eye.getX(), eye.getY(), eye.getZ(),
                      ray.getX(), ray.getY(), ray.getZ(),
                      squareIntersect.curveX * SQUARE_SIZE + 200,
                      -150,
                      squareIntersect.curveZ * SQUARE_SIZE + 200,
                      150, sphereIntersect);
                  if (sphereIntersect.intersected) {
                    double red = 0x40;
                    double green = 0x2c;
                    double blue = 0x04;
                    double lx = LIGHT_X - sphereIntersect.x;
                    double ly = LIGHT_Y - sphereIntersect.y;
                    double lz = LIGHT_Z - sphereIntersect.z;
                    double mag = magnitude(lx, ly, lz);
                    lx /= mag;
                    ly /= mag;
                    lz /= mag;
                    double nx = sphereIntersect.normalX;
                    double ny = sphereIntersect.normalY;
                    double nz = sphereIntersect.normalZ;
                    mag = magnitude(nx, ny, nz);
                    nx /= mag;
                    ny /= mag;
                    nz /= mag;
                    double diffuse = Math.max(0, lx * nx + ly * ny + lz * nz);
                    double rx = 2 * diffuse * nx - lx;
                    double ry = 2 * diffuse * ny - ly;
                    double rz = 2 * diffuse * nz - lz;
                    double vx = -ray.getX();
                    double vy = -ray.getY();
                    double vz = -ray.getZ();
                    mag = magnitude(vx, vy, vz);
                    vx /= mag;
                    vy /= mag;
                    vz /= mag;
                    double specular = Math.pow(Math.max(0,
                        rx * vx + ry * vy + rz * vz),
                        75);
                    red += diffuse * 0xB5 + specular * 255;
                    green += diffuse * 0x82 + specular * 255;
                    blue += diffuse * 0x05 + specular * 255;

                    pixels[y][x][0] += Math.min(255, red);
                    pixels[y][x][1] += Math.min(255, green);
                    pixels[y][x][2] += Math.min(255, blue);
                    continue outter;
                  }
                }
                findSquareIntersect(squareIntersect);
              }
            }*/

            t = -eye.getY() / ray.getY();
            if (t > 0) {
              double surfaceX = eye.getX() + t * ray.getX();
              double surfaceZ = eye.getZ() + t * ray.getZ();
              int curveX = (int) Math.floor(surfaceX / SQUARE_SIZE);
              int curveZ = (int) Math.floor(surfaceZ / SQUARE_SIZE);

              if (curveX >= 0 && curveX < 766 && curveZ >= 0 && curveZ < 766
                      && hilbertCurve[curveZ][curveX] > 0) {
                /*findSphereIntersect(
                    surfaceX, 0, surfaceZ,
                    LIGHT_X - surfaceX, LIGHT_Y, LIGHT_Z - surfaceZ,
                    SQUARE_SIZE * curveX + 200, -150, SQUARE_SIZE * curveZ + 200,
                    150, sphereIntersect);
                if (!sphereIntersect.intersected) {
                  curveX--;
                  if (curveX >= 0 && curveX < 766 && curveZ >= 0 && curveZ < 766
                      && hilbertCurve[curveZ][curveX] > 0) {
                    findSphereIntersect(
                        surfaceX, 0, surfaceZ,
                        LIGHT_X - surfaceX, LIGHT_Y, LIGHT_Z - surfaceZ,
                        SQUARE_SIZE * curveX + 200, -150, SQUARE_SIZE * curveZ + 200,
                        150, sphereIntersect);
                  }
                  curveX++;
                }
                if (!sphereIntersect.intersected) {
                  curveX++;
                  if (curveX >= 0 && curveX < 766 && curveZ >= 0 && curveZ < 766
                      && hilbertCurve[curveZ][curveX] > 0) {
                    findSphereIntersect(
                        surfaceX, 0, surfaceZ,
                        LIGHT_X - surfaceX, LIGHT_Y, LIGHT_Z - surfaceZ,
                        SQUARE_SIZE * curveX + 200, -150, SQUARE_SIZE * curveZ + 200,
                        150, sphereIntersect);
                  }
                  curveX--;
                }
                if (!sphereIntersect.intersected) {
                  curveZ--;
                  if (curveX >= 0 && curveX < 766 && curveZ >= 0 && curveZ < 766
                      && hilbertCurve[curveZ][curveX] > 0) {
                    findSphereIntersect(
                        surfaceX, 0, surfaceZ,
                        LIGHT_X - surfaceX, LIGHT_Y, LIGHT_Z - surfaceZ,
                        SQUARE_SIZE * curveX + 200, -150, SQUARE_SIZE * curveZ + 200,
                        150, sphereIntersect);
                  }
                  curveZ++;
                }
                if (!sphereIntersect.intersected) {
                  curveZ++;
                  if (curveX >= 0 && curveX < 766 && curveZ >= 0 && curveZ < 766
                      && hilbertCurve[curveZ][curveX] > 0) {
                    findSphereIntersect(
                        surfaceX, 0, surfaceZ,
                        LIGHT_X - surfaceX, LIGHT_Y, LIGHT_Z - surfaceZ,
                        SQUARE_SIZE * curveX + 200, -150, SQUARE_SIZE * curveZ + 200,
                        150, sphereIntersect);
                  }
                  curveZ--;
                }*/
                /*if (sphereIntersect.intersected) {
                  pixels[y][x][0] += 0x45;
                  pixels[y][x][1] += 0x45;
                  pixels[y][x][2] += 0x45;
                } else {
                  pixels[y][x][0] += 0xFD;
                  pixels[y][x][1] += 0xFD;
                  pixels[y][x][2] += 0xFD;
                }*/
                if (hilbertCurve[curveZ][curveX] == ((HilbertCurve.LEFT << 3) | HilbertCurve.DOWN)) {
                  pixels[y][x][0] += 0xFF;
                  pixels[y][x][1] += 0;
                  pixels[y][x][2] += 0;
                } else {
                  pixels[y][x][0] += 0xFD;
                  pixels[y][x][1] += 0xFD;
                  pixels[y][x][2] += 0xFD;
                }
              } else if (curveX > -10 && curveX < 776 && curveZ > -10 && curveZ < 776) {

                squareIntersect.x = surfaceX;
                squareIntersect.y = 0;
                squareIntersect.z = surfaceZ;
                squareIntersect.vx = ray.getX();
                squareIntersect.vy = ray.getY();
                squareIntersect.vz = ray.getZ();
                squareIntersect.curveX = curveX;
                squareIntersect.curveZ = curveZ;

                while (squareIntersect.curveX > -10
                    && squareIntersect.curveX < 776
                    && squareIntersect.curveZ > -10
                    && squareIntersect.curveZ < 776
                    && squareIntersect.y < GROUND_Y) {
                  findSquareIntersect(squareIntersect);
                  if (squareIntersect.curveX >= 0
                      && squareIntersect.curveX < 766
                      && squareIntersect.curveZ >= 0
                      && squareIntersect.curveZ < 766
                      && hilbertCurve[squareIntersect.curveZ][squareIntersect.curveX] > 0) {

                    double wallRed = 0x46;
                    double wallGreen = 0x46;
                    double wallBlue = 0x46;
                    double wallX = squareIntersect.x;
                    double wallY = squareIntersect.y;
                    double wallZ = squareIntersect.z;
                    double wallNormalX = squareIntersect.normalX;
                    double wallNormalY = 0;
                    double wallNormalZ = squareIntersect.normalZ;
                    double lightX = LIGHT_X - wallX;
                    double lightY = LIGHT_Y - wallY;
                    double lightZ = LIGHT_Z - wallZ;
                    double lightMag = magnitude(lightX, lightY, lightZ);
                    lightX /= lightMag;
                    lightY /= lightMag;
                    lightZ /= lightMag;

                    double dot = dot(wallNormalX, wallNormalY, wallNormalZ,
                            lightX, lightY, lightZ);
                    if (dot > 0) {
                      squareIntersect.vx = lightX;
                      squareIntersect.vy = lightY;
                      squareIntersect.vz = lightZ;
                      squareIntersect.curveX = squareIntersect.lastCurveX;
                      squareIntersect.curveZ = squareIntersect.lastCurveZ;
                      while (squareIntersect.curveX > -10
                          && squareIntersect.curveX < 776
                          && squareIntersect.curveZ > -10
                          && squareIntersect.curveZ < 776
                          && squareIntersect.y > 0) {
                        findSquareIntersect(squareIntersect);
                        if (squareIntersect.curveX >= 0
                            && squareIntersect.curveX < 766
                            && squareIntersect.curveZ >= 0
                            && squareIntersect.curveZ < 766
                            && hilbertCurve[squareIntersect.curveZ][squareIntersect.curveX] > 0) {
                          break;
                        }
                      }

                      if (squareIntersect.y < 0) {
                        wallBlue = wallGreen = wallRed 
                            = Math.min(255, dot * 200);
                      }
                    }

                    pixels[y][x][0] += wallRed;
                    pixels[y][x][1] += wallGreen;
                    pixels[y][x][2] += wallBlue;

                    break;
                  }
                }
              }
            }
          }
        }
      }
    }

    saveImage("E:\\render.png", pixels);
  }

  private Vector _origin = new Vector();
  private Vector _ray = new Vector();
  private Vector _point = new Vector();
  private Vector _pointN = new Vector();
  private double[] times = new double[2];
  private Vector _bestPoint = new Vector();
  private Vector _bestPointN = new Vector();
  private Vector _vec1 = new Vector();
  private Vector _vec2 = new Vector();
  private void findSolidIntersect(
      Vector origin, Vector ray, Solid solid, SolidIntersect solidIntersect,
      TransformPair transformPair, boolean reverseTransform) {

    solidIntersect.intersected = false;
    origin.copyTo(_origin);
    ray.copyTo(_ray);
    origin = _origin;
    ray = _ray;
    ray.addVector(origin);
    transformPair.inverseTransform.multiplyVector(origin);
    transformPair.inverseTransform.multiplyVector(ray);
    solid.inverseTransform.multiplyVector(origin);
    solid.inverseTransform.multiplyVector(ray);
    ray.subtractVector(origin);

    if (solid.isSphere) {
      double a = ray.getX() * ray.getX() + ray.getY() * ray.getY()
          + ray.getZ() * ray.getZ();
      double b = 2.0 * (ray.getX() * origin.getX() + ray.getY() * origin.getY()
          + ray.getZ() * origin.getZ());
      double c = origin.getX() * origin.getX() + origin.getY() * origin.getY()
          + origin.getZ() * origin.getZ() - solid.radius * solid.radius;
      double d = b * b - 4 * a * c;
      if (d < 0) {
        return;
      }
      d = Math.sqrt(d);
      a *= 2;
      double t1 = (-b + d) / a;
      double t2 = (-b - d) / a;
      if (t1 > t2) {
        times[0] = t2;
        times[1] = t1;
      } else {
        times[0] = t1;
        times[1] = t2;
      }
      for(int i = 0; i < 2; i++) {
        double t = times[i];
        if (t > 0) {
          double x = origin.getX() + ray.getX() * t;
          double y = origin.getY() + ray.getY() * t;
          double z = origin.getZ() + ray.getZ() * t;
          double nx = x;
          double ny = y;
          double nz = z;
          if (nx * ray.getX() + ny * ray.getY() + nz * ray.getZ() < 0) {
            solidIntersect.intersected = true;
            if (reverseTransform) {
              _point.setX(x);
              _point.setY(y);
              _point.setZ(z);
              _pointN.setX(2.0 * x);
              _pointN.setY(2.0 * y);
              _pointN.setZ(2.0 * z);
              solid.transform.multiplyVector(_point);
              solid.transform.multiplyVector(_pointN);
              transformPair.transform.multiplyVector(_point);
              transformPair.transform.multiplyVector(_pointN);
              _pointN.subtractVector(_point);
              _pointN.normalize();
              solidIntersect.point.copyFrom(_point);
              solidIntersect.normal.copyFrom(_pointN);
            }
            return;
          }
        }
      }
    } else {
      boolean intersect = false;
      double bestTime = Double.MAX_VALUE;
      double R2 = solid.radius * solid.radius;

      double t = (solid.length - origin.getX()) / ray.getX();
      if (t > 0 && ray.getX() < 0) {        
        _vec1.copyFrom(ray);
        _vec1.scale(t);
        _vec1.addVector(origin);
        _vec2.setX(solid.length);
        _vec2.setY(0);
        _vec2.setZ(0);
        if (Vector.distance2(_vec1, _vec2) <= R2) {
          _bestPoint.copyFrom(_vec1);
          _bestPointN.copyFrom(_bestPoint);
          _bestPointN.setX(_bestPointN.getX() + 1);
          bestTime = t;
          intersect = true;
        }
      }

      t = (-solid.length - origin.getX()) / ray.getX();
      if (t > 0 && ray.getX() > 0 && t < bestTime) {
        _vec1.copyFrom(ray);
        _vec1.scale(t);
        _vec1.addVector(origin);
        _vec2.setX(-solid.length);
        _vec2.setY(0);
        _vec2.setZ(0);
        if (Vector.distance2(_vec1, _vec2) <= R2) {
          _bestPoint.copyFrom(_vec1);
          _bestPointN.copyFrom(_bestPoint);
          _bestPointN.setX(_bestPointN.getX() - 1);
          bestTime = t;
          intersect = true;
        }
      }

      double a = ray.getY() * ray.getY() + ray.getZ() * ray.getZ();
      double b = 2.0 * (origin.getY() * ray.getY() + origin.getZ() * ray.getZ());
      double c = origin.getY() * origin.getY() + origin.getZ() * origin.getZ()
          - solid.radius * solid.radius;
      double d = b * b - 4 * a * c;
      if (d >= 0) {
        d = Math.sqrt(d);
        a *= 2;

        t = (-b + d) / a;
        if (t > 0 && t < bestTime) {
          _vec1.copyFrom(ray);
          _vec1.scale(t);
          _vec1.addVector(origin);
          if (_vec1.getX() <= solid.length && _vec1.getX() >= -solid.length) {
            _vec2.copyFrom(_vec1);
            _vec2.setX(0);
            if (_vec2.dot(ray) < 0) {
              _bestPoint.copyFrom(_vec1);
              _bestPointN.copyFrom(_bestPoint);
              _bestPointN.addVector(_vec2);
              bestTime = t;
              intersect = true;
            }
          }
        }

        t = (-b - d) / a;
        if (t > 0 && t < bestTime) {
          _vec1.copyFrom(ray);
          _vec1.scale(t);
          _vec1.addVector(origin);
          if (_vec1.getX() <= solid.length && _vec1.getX() >= -solid.length) {
            _vec2.copyFrom(_vec1);
            _vec2.setX(0);
            if (_vec2.dot(ray) < 0) {
              _bestPoint.copyFrom(_vec1);
              _bestPointN.copyFrom(_bestPoint);
              _bestPointN.addVector(_vec2);
              bestTime = t;
              intersect = true;
            }
          }
        }
      }

      if (intersect) {
        solidIntersect.intersected = true;
        if (reverseTransform) {
          solid.transform.multiplyVector(_bestPoint);
          solid.transform.multiplyVector(_bestPointN);
          transformPair.transform.multiplyVector(_bestPoint);
          transformPair.transform.multiplyVector(_bestPointN);
          _bestPointN.subtractVector(_bestPoint);
          _bestPointN.normalize();
          solidIntersect.point.copyFrom(_bestPoint);
          solidIntersect.normal.copyFrom(_bestPointN);
        }
        return;
      }
    }
  }

  private void findSphereIntersect(
      double ex, double ey, double ez,
      double dx, double dy, double dz,
      double cx, double cy, double cz,
      double r, SphereIntersect sphereIntersect) {
    sphereIntersect.intersected = false;
    double a = dx * dx + dy * dy + dz * dz;
    double vx = ex - cx;
    double vy = ey - cy;
    double vz = ez - cz;
    double b = 2.0 * (dx * vx + dy * vy + dz * vz);
    double c = vx * vx + vy * vy + vz * vz - r * r;
    double d = b * b - 4 * a * c;
    if (d < 0) {
      return;
    }
    d = Math.sqrt(d);
    a *= 2;
    double t1 = (-b + d) / a;
    double t2 = (-b - d) / a;
    if (t1 > t2) {
      double temp = t1;
      t1 = t2;
      t2 = temp;
    }
    if (t1 > 0) {
      double x = ex + dx * t1;
      double y = ey + dy * t1;
      double z = ez + dz * t1;
      double nx = x - cx;
      double ny = y - cy;
      double nz = z - cz;
      if (nx * dx + ny * dy + nz * dz < 0) {
        sphereIntersect.intersected = true;
        sphereIntersect.x = x;
        sphereIntersect.y = y;
        sphereIntersect.z = z;
        sphereIntersect.normalX = nx;
        sphereIntersect.normalY = ny;
        sphereIntersect.normalZ = nz;
        return;
      }
    }
    if (t2 > 0) {
      double x = ex + dx * t2;
      double y = ey + dy * t2;
      double z = ez + dz * t2;
      double nx = x - cx;
      double ny = y - cy;
      double nz = z - cz;
      if (nx * dx + ny * dy + nz * dz < 0) {
        sphereIntersect.intersected = true;
        sphereIntersect.x = x;
        sphereIntersect.y = y;
        sphereIntersect.z = z;
        sphereIntersect.normalX = nx;
        sphereIntersect.normalY = ny;
        sphereIntersect.normalZ = nz;
        return;
      }
    }
  }

  private double dot(
      double vx, double vy, double vz,
      double ux, double uy, double uz) {
    return vx * ux + vy * uy + vz * uz;
  }

  private double magnitude(double x, double y, double z) {
    return Math.sqrt(x * x + y * y + z * z);
  }

  private void findSquareIntersect(SquareIntersect squareIntersect) {

    double x = squareIntersect.x;
    double y = squareIntersect.y;
    double z = squareIntersect.z;
    double vx = squareIntersect.vx;
    double vy = squareIntersect.vy;
    double vz = squareIntersect.vz;
    int curveX = squareIntersect.curveX;
    int curveZ = squareIntersect.curveZ;
    squareIntersect.lastCurveX = curveX;
    squareIntersect.lastCurveZ = curveZ;

    double squareX = SQUARE_SIZE * curveX;
    double squareZ = SQUARE_SIZE * curveZ;

    if (vx < 0) {
      double t = (squareX - x) / vx;
      double newZ = z + t * vz;
      if (newZ >= squareZ && newZ <= squareZ + SQUARE_SIZE) {
        squareIntersect.curveX = curveX - 1;
        squareIntersect.curveZ = curveZ;
        squareIntersect.t = t;
        squareIntersect.x = squareX;
        squareIntersect.y = y + t * vy;
        squareIntersect.z = newZ;
        squareIntersect.normalX = 1;
        squareIntersect.normalZ = 0;
        return;
      }
    }

    if (vx > 0) {
      double t = (squareX + SQUARE_SIZE - x) / vx;
      double newZ = z + t * vz;
      if (newZ >= squareZ && newZ <= squareZ + SQUARE_SIZE) {
        squareIntersect.curveX = curveX + 1;
        squareIntersect.curveZ = curveZ;
        squareIntersect.t = t;
        squareIntersect.x = squareX + SQUARE_SIZE;
        squareIntersect.y = y + t * vy;
        squareIntersect.z = newZ;
        squareIntersect.normalX = -1;
        squareIntersect.normalZ = 0;
        return;
      }
    }

    if (vz < 0) {
      double t = (squareZ - z) / vz;
      double newX = x + t * vx;
      if (newX >= squareX && newX <= squareX + SQUARE_SIZE) {
        squareIntersect.curveX = curveX;
        squareIntersect.curveZ = curveZ - 1;
        squareIntersect.t = t;
        squareIntersect.x = newX;
        squareIntersect.y = y + t * vy;
        squareIntersect.z = squareZ;
        squareIntersect.normalX = 0;
        squareIntersect.normalZ = 1;
        return;
      }
    }

    if (vz > 0) {
      double t = (squareZ + SQUARE_SIZE - z) / vz;
      double newX = x + t * vx;
      if (newX >= squareX && newX <= squareX + SQUARE_SIZE) {
        squareIntersect.curveX = curveX;
        squareIntersect.curveZ = curveZ + 1;
        squareIntersect.t = t;
        squareIntersect.x = newX;
        squareIntersect.y = y + t * vy;
        squareIntersect.z = squareZ + SQUARE_SIZE;
        squareIntersect.normalX = 0;
        squareIntersect.normalZ = -1;
        return;
      }
    }

    throw new RuntimeException("Unhandled case: "
        + " x = " + x + ", "
        + " y = " + y + ", "
        + " z = " + z + ", "
        + " vx = " + vx + ", "
        + " vy = " + vy + ", "
        + " vz = " + vz + ", "
        + " curveX = " + curveX + ", "
        + " curveZ = " + curveZ);
  }

  private void saveImage(String fileName, double[][][] pixels) throws Throwable {
    double scale = SAMPLES * SAMPLES;
    int[] data = new int[WIDTH * HEIGHT];
    for(int i = 0, k = 0; i < HEIGHT; i++) {
      for(int j = 0; j < WIDTH; j++, k++) {
        int red = (int)(pixels[i][j][0] / scale);
        int green = (int)(pixels[i][j][1] / scale);
        int blue = (int)(pixels[i][j][2] / scale);
        data[k] = (red << 16) | (green << 8) | blue;
      }
    }
    BufferedImage image = new BufferedImage(
        WIDTH, HEIGHT, BufferedImage.TYPE_INT_RGB);
    image.setRGB(0, 0, WIDTH, HEIGHT, data, 0, WIDTH);
    ImageIO.write(image, "png", new File(fileName));
  }

  public static void main(String[] args) throws Throwable  {
    if (args.length != 5) {
      System.out.println("args: [output path] [begin scene (1--6)] [end scene (1--6)] [begin frame (0--59)] [end frame (0--59)]");
      return;
    }
    String path = args[0];
    int beginScene = Integer.parseInt(args[1]);
    int endScene = Integer.parseInt(args[2]);
    int beginFrame = Integer.parseInt(args[3]);
    int endFrame = Integer.parseInt(args[4]);
    Main main = new Main();
    for(int scene = beginScene; scene <= endScene; scene++) {
      for(int frame = beginFrame; frame <= endFrame; frame++) {
        main.render(scene, frame, path);
      }
    }
    
    //new Main().render(5, 30, "C:\\megacycles\\render2");
  }
}
