package megacycles;

public class Solid {
  public AffineTransform transform = AffineTransform.getIdentityTransform();
  public AffineTransform inverseTransform = AffineTransform.getIdentityTransform();
  public boolean isSphere = true;
  public double radius = 1;
  public double length = 1;
  public double ambientRed = 0x40;
  public double ambientGreen = 0x2c;
  public double ambientBlue = 0x04;
  public double diffuseRed = 0xB5;
  public double diffuseGreen = 0x82;
  public double diffuseBlue = 0x05;

  public Solid translate(double tx, double ty, double tz) {
    transform = AffineTransform.getTranslationTransform(tx, ty, tz).multiply(transform);
    inverseTransform = inverseTransform.multiply(AffineTransform.getTranslationTransform(-tx, -ty, -tz));
    return this;
  }

  public Solid scale(double sx, double sy, double sz) {
    transform = AffineTransform.getScaleTransform(sx, sy, sz).multiply(transform);
    inverseTransform = inverseTransform.multiply(AffineTransform.getScaleTransform(1.0 / sx, 1.0 / sy, 1.0 / sz));
    return this;
  }

  public Solid rotateAroundX(double angle, double y, double z) {
    return translate(0, -y, -z).rotateAroundX(angle).translate(0, y, z);
  }

  public Solid rotateAroundX(double angle) {
    transform = AffineTransform.getRotateAroundXTransform(angle).multiply(transform);
    inverseTransform = inverseTransform.multiply(AffineTransform.getRotateAroundXTransform(-angle));
    return this;
  }

  public Solid rotateAroundY(double angle, double x, double z) {
    return translate(-x, 0, -z).rotateAroundY(angle).translate(x, 0, z);
  }

  public Solid rotateAroundY(double angle) {
    transform = AffineTransform.getRotateAroundYTransform(angle).multiply(transform);
    inverseTransform = inverseTransform.multiply(AffineTransform.getRotateAroundYTransform(-angle));
    return this;
  }

  public Solid rotateAroundZ(double angle, double x, double y) {
    return translate(-x, -y, 0).rotateAroundZ(angle).translate(x, y, 0);
  }

  public Solid rotateAroundZ(double angle) {
    transform = AffineTransform.getRotateAroundZTransform(angle).multiply(transform);
    inverseTransform = inverseTransform.multiply(AffineTransform.getRotateAroundZTransform(-angle));
    return this;
  }
}
