package megacycles;

public class TransformPair {
  public AffineTransform transform = AffineTransform.getIdentityTransform();
  public AffineTransform inverseTransform = AffineTransform.getIdentityTransform();

  public void offset(double x, double y, double z) {
    transform.translate(x, y, z, false);
    inverseTransform.translate(-x, -y, -z, true);
  }

  public void undoOffset() {
    transform.undoTranslate();
    inverseTransform.undoTranslate();
  }

  public void setTranslation(double x, double y, double z) {
    transform.setTranslation(x, y, z);
    inverseTransform.setTranslation(-x, -y, -z);
  }

  public TransformPair translate(double tx, double ty, double tz) {
    transform = AffineTransform.getTranslationTransform(tx, ty, tz).multiply(transform);
    inverseTransform = inverseTransform.multiply(AffineTransform.getTranslationTransform(-tx, -ty, -tz));
    return this;
  }

  public TransformPair scale(double sx, double sy, double sz) {
    transform = AffineTransform.getScaleTransform(sx, sy, sz).multiply(transform);
    inverseTransform = inverseTransform.multiply(AffineTransform.getScaleTransform(1.0 / sx, 1.0 / sy, 1.0 / sz));
    return this;
  }

  public TransformPair rotateAroundX(double angle, double y, double z) {
    return translate(0, -y, -z).rotateAroundX(angle).translate(0, y, z);
  }

  public TransformPair rotateAroundX(double angle) {
    transform = AffineTransform.getRotateAroundXTransform(angle).multiply(transform);
    inverseTransform = inverseTransform.multiply(AffineTransform.getRotateAroundXTransform(-angle));
    return this;
  }

  public TransformPair rotateAroundY(double angle, double x, double z) {
    return translate(-x, 0, -z).rotateAroundY(angle).translate(x, 0, z);
  }

  public TransformPair rotateAroundY(double angle) {
    transform = AffineTransform.getRotateAroundYTransform(angle).multiply(transform);
    inverseTransform = inverseTransform.multiply(AffineTransform.getRotateAroundYTransform(-angle));
    return this;
  }

  public TransformPair rotateAroundZ(double angle, double x, double y) {
    return translate(-x, -y, 0).rotateAroundZ(angle).translate(x, y, 0);
  }

  public TransformPair rotateAroundZ(double angle) {
    transform = AffineTransform.getRotateAroundZTransform(angle).multiply(transform);
    inverseTransform = inverseTransform.multiply(AffineTransform.getRotateAroundZTransform(-angle));
    return this;
  }
}
